;------------------------------------------------------------------
; This macro is used to align code at a larger alignment than the
; current segment alignment (the limit for the align directive).
; It will work only for items in the code segment and only if
; there is a code_base label at the start of the code segment.
;------------------------------------------------------------------

codealign MACRO alignment
    LOCAL location, aligned_location
    location = $ - code_base
    aligned_location = location + alignment AND NOT (alignment - 1)
    db (alignment + aligned_location - location) dup(90h)
ENDM

; -----------------------------------------------------------------
; These two macros perform the grunt work involved in measuring
; the processor clock-cycle count for a block of code. These
; macros must be used in pairs, and the block of code must be
; placed in between the ctr_begin and ctr_end macro calls.
; The ctr_end macro returns the lowest cycle count that occurs
; in any single pass through the block of code, corrected for
; the loop overhead, in EAX (actually in EDX:EAX, but the cycle
; count is unlikely to overflow EAX).
;
; These macros require a .586 or higher processor directive.
; -----------------------------------------------------------------

ctr_begin MACRO loopcount:REQ

    LOCAL refLabel,workLabel

    IFNDEF __ctr__stuff__defined__
        __ctr__stuff__defined__ equ <1>
        .data
            ALIGN 2           ;; Not optimal but max supported
            __ctr__qword__count__     dq 0
            __ctr__qword__overhead__  dq 0
            __ctr__loop__count__      dd 0
            __ctr__loop__counter__    dd 0
        .code
    ENDIF

    cli                       ;; Minimize interruptions

    mov DWORD PTR __ctr__qword__count__, -1
    mov DWORD PTR __ctr__qword__count__ + 4, -1
    mov DWORD PTR __ctr__qword__overhead__, -1
    mov DWORD PTR __ctr__qword__overhead__ + 4, -1

    mov __ctr__loop__count__, loopcount
    mov __ctr__loop__counter__, loopcount

    push ebx                  ;; Preserve EBX around CPUID
    xor eax, eax              ;; Warm up the CPUID instruction
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    pop ebx

    codealign 16              ;; Optimal loop alignment for P6

  refLabel:                   ;; Reference loop

    __ref__loop__label__ equ <refLabel>

    push ebx                  ;; Preserve EBX around CPUID
    xor eax, eax              ;; Use same CPUID input value for each call
    cpuid                     ;; Flush pipe & wait for pending ops to finish
    rdtsc                     ;; Read Time Stamp Counter
    pop ebx

    push edx                  ;; Preserve high-order 32 bits of start count
    push eax                  ;; Preserve low-order 32 bits of start count

    push ebx
    xor eax, eax
    cpuid                     ;; Make sure loop setup instructions finish
    pop ebx

    ;;;;;;;;;;

    push ebx
    xor eax, eax
    cpuid                     ;; Make sure loop instructions finish
    rdtsc                     ;; Read end count
    pop ebx

    pop ecx                   ;; Recover low-order 32 bits of start count
    sub eax, ecx              ;; Low-order 32 bits of loop count in EAX
    pop ecx                   ;; Recover high-order 32 bits of start count
    sbb edx, ecx              ;; High-order 32 bits of loop count in EDX
    ;;
    ;; Capture lowest cycle count that occurs in any loop.
    ;;
    cmp edx, DWORD PTR __ctr__qword__overhead__ + 4
    jne @F
    cmp eax, DWORD PTR __ctr__qword__overhead__
  @@:
    jnb @F
    mov DWORD PTR __ctr__qword__overhead__, eax
    mov DWORD PTR __ctr__qword__overhead__ + 4, edx
  @@:
    sub __ctr__loop__counter__, 1
    jnz __ref__loop__label__  ;; End of reference loop

    mov __ctr__loop__counter__, loopcount

    codealign 16              ;; Optimal loop alignment for P6

  workLabel:                  ;; Working loop

    __work__loop__label__ equ <workLabel>

    push ebx                  ;; Preserve EBX around CPUID
    xor eax, eax              ;; Use same CPUID input value for each call
    cpuid                     ;; Flush pipe & wait for pending ops to finish
    rdtsc                     ;; Read Time Stamp Counter
    pop ebx

    push edx                  ;; Preserve high-order 32 bits of start count
    push eax                  ;; Preserve low-order 32 bits of start count

    push ebx
    xor eax, eax
    cpuid                     ;; Make sure loop setup instructions finish
    pop ebx

ENDM

ctr_end MACRO

    push ebx
    xor eax, eax
    cpuid                     ;; Make sure loop instructions finish
    rdtsc                     ;; Read end count
    pop ebx

    pop ecx                   ;; Recover low-order 32 bits of start count
    sub eax, ecx              ;; Low-order 32 bits of loop count in EAX
    pop ecx                   ;; Recover high-order 32 bits of start count
    sbb edx, ecx              ;; High-order 32 bits of loop count in EDX
    ;;
    ;; Capture lowest cycle count that occurs in any loop.
    ;;
    cmp edx, DWORD PTR __ctr__qword__count__ + 4
    jne @F
    cmp eax, DWORD PTR __ctr__qword__count__
  @@:
    jnb @F
    mov DWORD PTR __ctr__qword__count__, eax
    mov DWORD PTR __ctr__qword__count__ + 4, edx
  @@:

    sub __ctr__loop__counter__, 1
    jnz __work__loop__label__  ;; End of working loop

    mov eax, DWORD PTR __ctr__qword__count__
    sub eax, DWORD PTR __ctr__qword__overhead__
    mov edx, DWORD PTR __ctr__qword__count__ + 4
    sbb edx, DWORD PTR __ctr__qword__overhead__ + 4

    sti

ENDM

;------------------------------------------------------------------
; The following is a minimal set of support macros and procedures,
; along with the necessary equates and data definitions.
;------------------------------------------------------------------

    ; -----------------------------------------------
    ; DOS handles for standard input/output devices.
    ; -----------------------------------------------

    HSTDIN   EQU 0
    HSTDOUT  EQU 1

    ; -----------------------------------------------------
    ; These prototypes are not actually necessary, but are
    ; included primarily to serve as a syntax reference.
    ; -----------------------------------------------------

    waitkey PROTO
    stdout PROTO c npsz:WORD
    dw2dec PROTO c src:DWORD, npdest:WORD
    get_tick_count PROTO c
    sleep PROTO c ticks:DWORD

    ; -----------------------------------------------------
    ; This is the MASM32 reparg macro with *no* changes.
    ;
    ; "This macro replaces quoted text with a DATA section
    ; OFFSET and returns it in ADDR "name" format."
    ; -----------------------------------------------------

    reparg MACRO arg
      LOCAL nustr
      quot SUBSTR <arg>, 1, 1
      IFIDN quot,<">          ;; if 1st char = "
        .data
          nustr db arg, 0     ;; write arg to .DATA section
        .code
        EXITM <ADDR nustr>    ;; append name to ADDR operator
      ELSE
        EXITM <arg>           ;; else return arg
      ENDIF
    ENDM

    ; ------------------------------------------------------
    ; This is the MASM32 chr$ macro with *no* changes.
    ;
    ; "This macro takes quoted string and ascii numbers in
    ; the same format as a .DATA section entry. It returns
    ; the offset in the .DATA section of the data contained
    ; in its parameters."
    ; ------------------------------------------------------

    chr$ MACRO any_text:VARARG
      LOCAL txtname
      .data
        txtname db any_text,0
      .code
      EXITM <OFFSET txtname>
    ENDM

    ; ------------------------------------------------------
    ; This is the MASM32 print macro with *no* changes.
    ; ------------------------------------------------------

    print MACRO arg1:REQ,varname:VARARG      ;; display zero terminated string
        invoke StdOut,reparg(arg1)
      IFNB <varname>
        invoke StdOut,chr$(varname)
      ENDIF
    ENDM

    dword$ MACRO src
      invoke dw2dec,src,ADDR outBuf
      EXITM <ADDR outBuf>
    ENDM

.data
    outbuf db 30 dup(0)
.code

code_base:

; -------------------------------------------------------------
; This proc returns the length of the specified
; null-terminated string in AX.
; -------------------------------------------------------------

szlen proc c uses bx npsz:WORD
    mov bx, npsz
  runLoop:
    cmp BYTE PTR [bx], 0      ; Compare bytes until find null
    jz  @F
    inc bx
    jmp runLoop
  @@:
    mov ax, bx                ; Calc length and return in AX
    sub ax, npsz
    ret
szlen endp

; -------------------------------------------------------------
; This proc calls the Write File or Device function to
; output the specified null-terminated string to STDOUT.
; -------------------------------------------------------------

stdout proc c uses ax bx cx dx npsz:WORD
    invoke szlen, npsz
    mov cx, ax
    mov bx, HSTDOUT
    mov dx, npsz
    mov ah, 40h
    int 21h
    ret
stdout endp

; -------------------------------------------------------------
; This proc calls the BIOS Read Keyboard Status and Read
; Keyboard Input functions to flush the keyboard buffer,
; wait for a key press, and then flush the buffer again,
; leaving it empty. The BIOS scan code for the key pressed
; is returned in AH and the character code in AL.
; -------------------------------------------------------------

waitkey proc c
  @@:
    mov ah, 1
    int 16h                   ; Check for key press
    jz  @F                    ; Jump if buffer empty
    mov ah, 0
    int 16h                   ; Get key
    jmp @B                    ; Repeat
  @@:
    mov ah, 1
    int 16h                   ; Check for key press
    jz  @B                    ; Repeat
    mov ah, 0
    int 16h                   ; Get key
    push ax                   ; Preserve it
  @@:
    mov ah, 1
    int 16h                   ; Check for key press
    jz  @F                    ; Jump if buffer empty
    mov ah, 0
    int 16h                   ; Get key
    jmp @B                    ; Repeat
  @@:
    pop ax                    ; Recover key
    ret
waitkey endp

; -------------------------------------------------------------
; This proc converts a dword to an signed null-terminated
; decimal string and stores the result (11 characters max)
; in the specified buffer.
;
; This proc cannot correctly handle values that are outside
; the range -2147483648 to 2147483647.
; -------------------------------------------------------------

dw2dec proc c uses ax bx cx dx di src:DWORD, npdest:WORD
    mov eax, src
    mov di, npdest
    test eax, eax             ; Test for negative value
    jns @F
    neg eax                   ; Make positive if negative
    mov BYTE PTR[di], '-'     ; Store sign in buffer
    inc di
  @@:
    mov ebx, 10               ; Init divisor
    xor cx, cx                ; Zero digit counter
  divLoop:
    xor edx, edx              ; Zero MSDW of dividend
    div ebx                   ; Divide value by 10
    push dx                   ; Push remainder
    inc cx                    ; Increment digit count
    or  eax, eax              ; Repeat if quotient > 0
    jnz divLoop
  storeLoop:
    pop ax                    ; Pop digit off stack
    add al, '0'               ; Convert to ascii
    mov [di], al              ; Store in buffer
    inc di                    ; Adjust DI for next character
    loop storeLoop            ; Repeat CX times
    mov BYTE PTR [di], 0      ; Append trailing null
    ret
dw2dec endp

; -------------------------------------------------------------
; This proc returns the BIOS maintained count of timer ticks
; since midnight in EAX. The timer ticks are generated by
; PIT channel 0, normally programed by the BIOS to generate
; 1193182/65536 = 18.207 ticks per second.
; -------------------------------------------------------------

get_tick_count proc uses es bx
    mov ax, 40h               ; BIOS data area segment address
    mov es, ax
    mov bx, 6ch
    mov eax, es:[bx]
    ret
get_tick_count endp

; -------------------------------------------------------------
; This proc delays for ticks * 55ms before returning.
; -------------------------------------------------------------

sleep proc ticks:DWORD
    call get_tick_count
    mov edx, eax
    add edx, ticks
  @@:
    call get_tick_count
    cmp eax, edx
    jb  @B
    ret
sleep endp
